import ta
from ta.momentum import *
from ta.trend import *
from ta.volatility import *


class Indicators:
    _df = None

    @property
    def indicators_df(self):
        return self._df

    @indicators_df.setter
    def indicators_df(self, df):
        self._df = df

    def BollingerBands(self, **kwargs):
        return BollingerBands(self.indicators_df["Close"], **kwargs)

    def SMA(self, **kwargs):
        return SMAIndicator(self.indicators_df["Close"], **kwargs).sma_indicator()

    def EMA(self, **kwargs):
        return EMAIndicator(self.indicators_df["Close"], **kwargs).ema_indicator()

    def MACD(self, **kwargs):
        return MACD(self.indicators_df["Close"], **kwargs)

    def RSI(self, **kwargs):
        return RSIIndicator(self.indicators_df["Close"], **kwargs).rsi()

    def ATR(self, **kwargs):
        return AverageTrueRange(
            self.indicators_df["High"],
            self.indicators_df["Low"],
            self.indicators_df["Close"],
            **kwargs
        ).average_true_range()



from random import randint

from python_trader.df_tools import DFTools


class Trader(DFTools):
    def __init__(self):

        self._active_columns = [
            "id",
            "open_time",
            "type",
            "size",
            "asset",
            "open_price",
            "sl",
            "tp",
            "swap",
            "commission",
            "single_profit",
        ]

        self._history_columns = [
            "id",
            "open_time",
            "type",
            "size",
            "asset",
            "open_price",
            "sl",
            "tp",
            "close_time",
            "close_price",
            "single_profit",
            "comulative_profit",
        ]

        self._pending_columns = [
            "id",
            "open_time",
            "type",
            "size",
            "asset",
            "open_price",
            "sl",
            "tp",
            "swap",
            "commission",
            "single_profit",
        ]

        super().__init__()
        self._current_bar = None
        self._current_asset = None
        self._register_data = None
        self._balance = None

    @property
    def register_data(self):
        return self._register_data

    @register_data.setter
    def register_data(self, data):
        self._register_data = data

    @property
    def current_bar(self):
        if self._current_bar.empty:
            print("There is not candle info")
            return False
        return self._current_bar

    @current_bar.setter
    def current_bar(self, bar):
        self._current_bar = bar

    @property
    def current_asset(self):
        return self._current_asset

    @current_asset.setter
    def current_asset(self, asset):
        self._current_asset = asset

    @property
    def balance(self):
        return self._balance

    @balance.setter
    def balance(self, balance):
        self._balance = balance

    def _common_position(self, type, **kwargs):
        dic = {}
        for item, val in kwargs.items():
            if item not in self._active_columns:
                raise Exception(f"Invalid argument {item}")
            dic[item] = val

        dic["type"] = type
        dic["open_time"] = self.current_bar["DateTime"]
        dic["asset"] = self.current_asset
        # assigning random id ... unlikely will be duplicated
        dic["id"] = self._make_rand_id()

        if "size" not in dic:
            dic["size"] = 0.01

        # checking if position is already taken
        search_row = self.search_column(self.active_positions, "id", dic["id"])
        if not search_row.empty:
            print(f"position already exists id:{dic['id']}")
            return False

        # doing different things on pendings
        if ("stop" in dic["type"]) or ("limit" in dic["type"]):
            if "open_price" not in dic:
                raise Exception("Pending orders require an valid open_price")

            self.pending_orders = self.new_dic_record(self.pending_orders, dic)
        # market execution
        else:
            dic["open_price"] = self.current_bar["Close"]
            # change time on pending order filling
            self.active_positions = self.new_dic_record(self.active_positions, dic)

        self._validate_order(dic)
        return True

    def _validate_order(self, dic):
        """ Validates position based on given tp, sl, open_price"""

        open_price = dic["open_price"]
        current_close = self.current_bar["Close"]
        type = dic["type"]

        if "tp" in dic:
            tp = dic["tp"]
        else:
            tp = False

        if "sl" in dic:
            sl = dic["sl"]
        else:
            sl = False

        if type in ["buy_stop", "sell_limit"] and not open_price > current_close:
            raise Exception(f"{type} requires higher open_price")
        elif type in ["sell_stop", "buy_limit"] and not open_price < current_close:
            raise Exception(f"{type} requires lower open_price")

        if "buy" in type:
            if tp and not tp > open_price:
                raise Exception(f"{type} requires tp to be greater than open_price")
            if sl and not sl < open_price:
                raise Exception(f"{type} requires sl to be lesser than open_price")

        elif "sell" in type:
            if tp and not tp < open_price:
                raise Exception(f"{type} requires tp to be lesser than open_price")
            if sl and not sl > open_price:
                raise Exception(f"{type} requires sl to be greater than open_price")
        return True

    def buy(self, **kwargs):
        return self._common_position("buy", **kwargs)

    def sell(self, **kwargs):
        return self._common_position("sell", **kwargs)

    def buy_stop(self, **kwargs):
        return self._common_position("buy_stop", **kwargs)

    def sell_stop(self, **kwargs):
        return self._common_position("sell_stop", **kwargs)

    def buy_limit(self, **kwargs):
        return self._common_position("buy_limit", **kwargs)

    def sell_limit(self, **kwargs):
        return self._common_position("sell_limit", **kwargs)

    def _remove_active_position(self, pos_df, lot_size=None):
        # removing position from current
        id = pos_df["id"]
        self.active_positions = self.remove_record(self.active_positions, id)
        # adding position to history
        self.history = self.new_df_record(self.history, pos_df)
        return True

    def _activate_pending_position(self, pos_df, lot_size=None):
        # removing position from current
        id = pos_df["id"]
        self.pending_orders = self.remove_record(self.pending_orders, id)
        # adding position to history
        self.active_positions = self.new_df_record(self.active_positions, pos_df)
        return True

    def _remove_pending_position(self, pos_df, lot_size=None):
        # removing position from current
        id = pos_df["id"]
        self.pending_orders = self.remove_record(self.pending_orders, id)
        # adding position to history
        self.history = self.new_df_record(self.history, pos_df)
        return

    def cancel_pending(self, id):
        row = self.search_column(self.pending_orders, "id", id)
        # no such position
        if row.empty:
            return
        self._remove_pending_position(row)

    def _check_actives(self):
        for index, position in self.active_positions.iterrows():
            if position["asset"] != self.current_asset:
                continue
            if "buy" in position["type"]:
                self._check_buy_position(position)
            elif "sell" in position["type"]:
                self._check_sell_position(position)

    def _check_pendings(self):
        for index, position in self.pending_orders.iterrows():
            if position["asset"] != self.current_asset:
                continue
            self._check_pending_position(position, position["type"])

    def _check_pending_position(self, pos_df, type):
        high = self.current_bar["High"]
        low = self.current_bar["Low"]
        open_price = pos_df["open_price"]

        if high > open_price and (type == "buy_stop" or type == "sell_limit"):
            self._activate_pending_position(pos_df)
        elif low < open_price and (type == "sell_stop" or type == "buy_limit"):
            self._activate_pending_position(pos_df)

    def _check_buy_position(self, pos_df):
        high = self.current_bar["High"]
        low = self.current_bar["Low"]
        pos_id = pos_df["id"]

        if high >= pos_df["tp"]:
            pos_df = self._finalize_position(
                pos_df, self._get_profit("tp", pos_df), pos_df["tp"]
            )
            self._remove_active_position(pos_df)
        elif low <= pos_df["sl"]:
            pos_df = self._finalize_position(
                pos_df, self._get_profit("sl", pos_df), pos_df["sl"]
            )
            pos_df["single_profit"] = self._get_profit("sl", pos_df)
            self._remove_active_position(pos_df)

    def _finalize_position(self, pos_df, profit, close_price):
        pos_df["close_price"] = close_price
        pos_df["close_time"] = self.current_bar["DateTime"]
        pos_df["single_profit"] = profit
        self.balance += profit
        return pos_df

    def _check_sell_position(self, pos_df):
        high = self.current_bar["High"]
        low = self.current_bar["Low"]
        pos_id = pos_df["id"]

        if low <= pos_df["tp"]:
            pos_df = self._finalize_position(
                pos_df, self._get_profit("tp", pos_df), pos_df["tp"]
            )
            self._remove_active_position(pos_df)
        elif high >= pos_df["sl"]:
            pos_df = self._finalize_position(
                pos_df, self._get_profit("sl", pos_df), pos_df["sl"]
            )
            self._remove_active_position(pos_df)

    def close_position(self, id):
        pos_df = self.search_column(self.active_positions, "id")
        if pos_df.empty:
            return False

        pos_df["single_profit"] = self._get_profit("manual", pos_df)
        self._remove_active_position(pos_df)
        return True

    def _get_profit(self, mode, position, value=True, spread=True):
        # getting spread
        if spread:
            spread = self.register_data[self.current_asset]["spread"]
        else:
            spread = 0

        if position.empty:
            return False

        if mode == "tp":
            profit = abs(position["tp"] - position["open_price"])
        if mode == "sl":
            profit = -abs(position["sl"] - position["open_price"])
        elif mode == "manual":
            profit = self.current_bar["Close"] - position["open_price"]

        # normalizing profit
        profit = int(profit / self.register_data[self.current_asset]["point_value"])

        # subtracting spread from final profit
        profit -= spread

        if value:
            profit *= position["size"]

        return float(profit)

    @staticmethod
    def _make_rand_id(min_int=1000000, max_int=9999999):
        return randint(a=min_int, b=max_int)



import pandas as pd


class DFTools:
    _active_columns = []
    _history_columns = []
    _pending_columns = []
    _static_columns = []

    def __init__(self):
        self._history_df = self.new_df(self._history_columns)
        self._active_df = self.new_df(self._active_columns)
        self._pending_df = self.new_df(self._pending_columns)
        self._static_df = self.new_df(self._static_columns)

    @property
    def active_positions(self):
        return self._active_df

    @active_positions.setter
    def active_positions(self, df):
        self._active_df = df

    @property
    def history(self):
        return self._history_df

    @history.setter
    def history(self, df):
        self._history_df = df

    @property
    def pending_orders(self):
        return self._pending_df

    @pending_orders.setter
    def pending_orders(self, df):
        self._pending_df = df

    @property
    def statics(self):
        return self._static_df

    @statics.setter
    def statics(self, df):
        self._static_df = df

    def new_df(self, col_list):
        return pd.DataFrame(columns=col_list)

    def new_dic_record(self, df, dic):
        df = df.append(dic, ignore_index=True)
        return df

    def new_df_record(self, df_src, df):
        df_src = df_src.append(df)
        return df_src

    def search_column(self, df, col, kw):
        row = df.loc[df[col] == kw]
        return row

    def remove_record(self, df, id):
        df = df[df.id != id]
        return df

    def df_len(self, df):
        return df.shape[0]

    def get_history(self):
        return self.history_df

    def col_values(self, df, col):
        return df[col]

    def get_row_index(self, df, row):
        return df.index[df["DateTime"] == row["DateTime"]].tolist()[0]

    def get_row_by_index(self, df, index):
        return df.iloc[[index]]

    def cumsum(self, df, col, new_col):
        series = self.col_values(df, col)
        df[new_col] = series.cumsum()
        return df

    def export_csv(self, df, path, **kwargs):
        df.to_csv(path, **kwargs)



import os

from setuptools import find_packages, setup

requirements_path = "requirements.txt"
with open("README.md", "r") as fh:
    long_description = fh.read()


def get_package_info():

    init_path = os.path.join(os.path.dirname(__file__), "python_trader", "__init__.py")
    info_dic = {
        "__version__": "",
        "__description__": "",
        "__author__": "",
        "__email__": "",
    }

    with open(init_path, "rt") as info:
        for line in info:
            for info in info_dic.keys():
                if line.startswith(info):
                    info_dic[info] = eval(line.split("=")[1])
                    continue
    return info_dic


def parse_requirements(filename):
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]


install_reqs = parse_requirements(requirements_path)
package_info = get_package_info()
setup(
    name="python_trader",
    version=package_info["__version__"],
    description=package_info["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/quitegreensky/python_trader",
    author=package_info["__author__"],
    author_email=package_info["__email__"],
    license="MIT",
    install_requires=parse_requirements(requirements_path),
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)



from datetime import datetime
from functools import partial

from python_trader.indicators import Indicators
from python_trader.trader import Trader


class Strategy(Trader, Indicators):
    def __init__(self):
        super().__init__()

    def test(
        self, register_data, from_date, to_date, init_balance, previous_bars_count=100
    ):

        self.previous_bars_count = previous_bars_count
        self.register_data = register_data
        self.from_date = self.parse_str_datetime(from_date)
        self.to_date = self.parse_str_datetime(to_date)
        self.balance = init_balance
        # iterating throught assets, first completes one asset then goes to next.
        for asset, data in self.register_data.items():
            data["historical_data"].apply(
                partial(self._backtest, asset, data["historical_data"]), axis=1
            )

        self._on_finish()
        self.on_finish()

    def _backtest(self, asset, all_bars, bar):
        if not self._check_date(bar["DateTime"], self.from_date, self.to_date):
            return False

        index = self.get_row_index(all_bars, bar)
        if index > self.previous_bars_count:
            bars_until_now = all_bars[1 + index - self.previous_bars_count : index + 1]
        else:
            bars_until_now = all_bars[1 : index + 1]

        self.indicators_df = bars_until_now
        self.current_bar = bar
        self.current_asset = asset
        self._do_checks()
        self.on_bar_close(asset, bars_until_now.to_dict("records"))

    def _check_date(self, date, from_dt, to_dt):
        if from_dt <= date <= to_dt:
            return True
        else:
            return False

    def on_bar_close(self, asset, bar):
        pass

    def on_finish(self):
        pass

    def _on_finish(self):

        # calculating statics
        self.history = self.cumsum(self.history, "single_profit", "comulative_profit")
        com_series = self.history["comulative_profit"]
        profit_series = self.history["single_profit"]
        winner_series = profit_series.where(profit_series.gt(0)).dropna()
        losser_series = profit_series.where(profit_series.lt(0)).dropna()

        max_comulative_profit = com_series.max()
        max_drawdown = com_series.min()
        max_single_winner = winner_series.max()
        min_single_winner = winner_series.min()
        max_single_losser = losser_series.min()
        min_single_losser = losser_series.max()
        sum_of_winners = winner_series.sum()
        sum_of_lossers = losser_series.sum()
        winners_count = winner_series.size
        lossers_count = losser_series.size
        total_trades = lossers_count + winners_count
        average_winner = sum_of_winners / winners_count
        average_losser = sum_of_lossers / lossers_count
        profit_factor = abs(average_winner / average_losser)
        winrate = winners_count / total_trades
        # there is not index -1 -> profit is 0
        try:
            final_profit = self.get_row_by_index(com_series, -1)
        except IndexError:
            final_profit = 0

        dic = {
            "max_comulative_profit": max_comulative_profit,
            "max_drawdown": max_drawdown,
            "max_single_winner": max_single_winner,
            "min_single_winner": min_single_winner,
            "max_single_losser": max_single_losser,
            "min_single_losser": min_single_losser,
            "sum_of_winners": sum_of_winners,
            "sum_of_lossers": sum_of_lossers,
            "winners_count": winners_count,
            "lossers_count": lossers_count,
            "total_trades": total_trades,
            "average_winner": average_winner,
            "average_losser": average_losser,
            "profit_factor": profit_factor,
            "final_profit": float(final_profit),
            "winrate": winrate,
            "final_balance": self.balance,
        }

        self.statics = self.new_dic_record(self.statics, dic)

    def _do_checks(self):
        self._check_actives()
        self._check_pendings()

    @staticmethod
    def parse_str_datetime(t):
        try:
            return datetime.strptime(t, r"%Y.%m.%d %H:%M")
        except ValueError as e:
            raise ValueError(e)

    def indicator_index(self, df, index):
        return float(self.get_row_by_index(df, index))



import os
import warnings

import pandas as pd


class CsvHandler:
    standard_cols = ["Date", "Time", "Open", "High", "Low", "Close", "Volume"]

    def __init__(self, path, csv_type="standard", **kwargs):
        self.df = self._open_csv(path, csv_type)

    def _open_csv(self, path, csv_type):
        if csv_type == "mt5":
            warnings.filterwarnings("ignore")
            kw = {
                "delimiter": r"\t",
                "usecols": [0, 1, 2, 3, 4, 5, 7],
                "skiprows": lambda x: x in [0],
                "names": self.standard_cols,
            }
        elif csv_type == "mt4":
            kw = {"names": self.standard_cols}
        elif csv_type == "standard":
            kw = {"index_col": 0}
        else:
            raise KeyError

        df = pd.read_csv(path, **kw)

        if csv_type == "mt5":
            # removing seconds
            try:
                df["Time"] = df["Time"].apply(lambda x: x.rsplit(":", 1)[0])
            # mt5 daily exports doesn't have time column!!
            except AttributeError:
                # trying to add Time Column with default value
                new = [x for x in self.standard_cols if x != "Time"]
                new.append("Spread")
                df = df.rename(columns={x: y for x, y in zip(self.standard_cols, new)})
                df["Time"] = "00:00"

        df = self._merge_datetime(df)
        return df

    @staticmethod
    def _merge_datetime(df):
        df["DateTime"] = ""
        df["DateTime"] = df.apply(lambda row: f"{row['Date']} {row['Time']}", axis=1)
        df["DateTime"] = pd.to_datetime(df["DateTime"], format=r"%Y.%m.%d %H:%M")
        return df

    def save_to_csv(self, path, df=None):
        if not df:
            df = self.df
        df.to_csv(path)

    def get_df(self):
        return self.df
